/*

 compat.h - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define OPEN_RONLY 0
#define OPEN_STATUS 2
#define OPEN_WONLY 4
#define OPEN_ERR 6

int openfile_compat() {
    char s[256];
    unsigned char status=OPEN_ERR;
    unsigned char channel,fp;
    unsigned int length=0;
    struct t_buffer *buffer;
    int r;

    if (arguments.mode == M_ETHERNET) {
        channel=ebuf[3] & 0x0f;
    } else {
        r=io_ops->getb();if (r < 0) return 1;
        channel=r & 0x0f;
    }
    buffer=&buff[channel];

    fp=0;
    while (fp<255) {
        if (arguments.mode == M_ETHERNET) {
            s[fp]=ebuf[fp+4];
        } else {
            r=io_ops->getb();if (r < 0) return 1;
            s[fp]=r;
        }
        if (!s[fp]) break;
        fp++;
    }
    s[fp]=0;

    alarm(0);

    {
        if (buffer->file) {
            fclose(buffer->file);
            buffer->file=NULL;
        }
        buffer->mode=CM_CLOSED;

        {
            char vessz,type[4];
            unsigned char mode;
            char lname[1000],keszpath[1020];
            struct dirent *ep;
            DIR *dp;
            int found=0;

            resolv_path(s,keszpath,0);
            convertfilename(s,',',&vessz,s,type,&mode);

            dp=opendir(keszpath[0]?keszpath:".");
            if (!dp) {
                errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",keszpath);
                if (arguments.mode == M_ETHERNET) {
                    if (sendto(sock,"\2\2",2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
                } else {
                    if (io_ops->sendb(2)) return 1;
                }
                return 0;
            }
            if (type[0]==0) strcpy(type,(channel==1)?"PRG":"*");

            while ((ep=readdir(dp)))
            {
                struct stat buf;
                DIRECTORY_ENTRY_T dirent;
                char lname2[2020];

                strcpy(lname2,keszpath);
                strncpy(lname,ep->d_name,999);
                strncat(lname2,ep->d_name,999);

                if (lstat(lname2,&buf)) continue;

                createdirentry(ep,&buf,&dirent);

                if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name,s)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                found=1;length=buf.st_size;
                break;
            }

            closedir(dp);
            status=OPEN_ERR;

            if (!found) {
                int f=0,g=0;
                while (s[f]) {
                    if (f<16) {
                        lname[f]=c64toascii[(unsigned char)s[f]];
                        f++;
                    }
                }
		if (!relaxfiletypes || memcmp(type,"PRG",3)) {
            	    lname[f++]=',';
            	    while (type[g]) {
                	if (g<3) {
                    	    lname[f]=c64toascii[(unsigned char)type[g]];
                    	    f++;g++;
                	}
            	    }
		}
                lname[f]=0;
            }

            strcat(keszpath,lname);

            fprintf(lf,"Open [%d]: \"%s\": ",channel,keszpath);

            if (!s[0]) {
                seterror(34,0);goto vege;
            }

            {
                if (!found) {
                    seterror(62,0);
                } else {
#ifndef WIN32
                    buffer->fd=open(keszpath, O_RDONLY, 0);
#else
                    buffer->fd=open(keszpath, O_RDONLY|O_BINARY, 0);
#endif
                    if (buffer->fd<0) errtochannel15(1); else {
                        status=OPEN_RONLY;//ok
                        buffer->file=fdopen(buffer->fd, "rb");
                        buffer->mode=CM_COMPAT;
                        buffer->filesize=length;
                    }
                }
            }
        }
    }
vege:
    if (arguments.mode == M_ETHERNET) {
        ebuf[0]=status;ebuf[1]=status;if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
    } else {
        if (io_ops->sendb(status)) return 1;
    }
    return 0;
}

int readfile_compat() {
    unsigned char channel,status=0;
    unsigned short int bytes,i;
    struct timeval start,end;
    unsigned long long l=0,l2;
    struct t_buffer *buffer;
    int r;

    if (arguments.mode == M_ETHERNET) {
        channel=ebuf[3] & 0x0f;
        buffer=&buff[channel];
        bytes=ebuf[4];
        if (!bytes) bytes=256;
    } else {
        r=io_ops->getb();if (r < 0) return 1;
        channel=r & 0x0f;
        buffer=&buff[channel];
        r=io_ops->getb();if (r < 0) return 1;
        bytes=r;
        r=io_ops->getb();if (r < 0) return 1;
        bytes|=r << 8;
        if (io_ops->getb() < 0) return 1;
    }
    fprintf(lf,"Read [%d] %d bytes\n",channel,bytes);
    if (arguments.mode != M_ETHERNET) {
        if (io_ops->sendb(0)) return 1;
    }
    if (buffer->mode!=CM_COMPAT) {
        if (arguments.mode == M_ETHERNET) {
            ebuf[0]=0;ebuf[1]=0;if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
        } else {
            if (io_ops->sendb(0)) return 1;
            if (io_ops->sendb(0)) return 1;
            if (io_ops->sendb(0)) return 1;
        }
        return 0;
    } else {
        unsigned long itt;

        if (!buffer->data) {
            buffer->data=malloc(65536);
            if (!buffer->data) exit(1);//out of memory?!
        }

        l=bytes=fread(buffer->data,1,bytes?bytes:65536,buffer->file);
        if (feof(buffer->file)) status=0x42; else
        {
            itt=ftell(buffer->file);
            fgetc(buffer->file);
            if (feof(buffer->file)) status=0x40;
            fseek(buffer->file,itt,SEEK_SET);
        }
    }
    if (arguments.mode == M_ETHERNET) {
        ebuf[0]=bytes>0;
        ebuf[1]=bytes;
        ebuf[(bytes & 0xfffe)+3]=status;
        ebuf[((bytes+1) & 0xfffe)+2]=status;
        ebuf[((bytes+1) & 0xfffe)+3]=status;
        for (i=0;i<bytes;i++) ebuf[1+bytes-i]=buffer->data[i];
        if (sendto(sock,ebuf,((bytes+1) & 0xfffe)+4, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
    } else {
        if (io_ops->sendb(~bytes)) return 1;
        if (io_ops->sendb(~bytes >> 8)) return 1;
        io_ops->flush();

        gettimeofday(&start,NULL);
        if (bytes) {
            if (io_ops->sendbytes(bytes, buffer->data)) return 1;
        } else {
            if (io_ops->sendbytes(32768, buffer->data)) return 1;
            if (io_ops->sendbytes(32768, buffer->data+32768)) return 1;
        }
        if (io_ops->sendb(status)) return 1;//fprintf(lf,"Status: %02X\n",status);
        io_ops->flush();
        gettimeofday(&end,NULL);
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        if (l2) {
            l=l*1000000/l2;
            fprintf(lf,"Read %d bytes at %lld bytes/seconds\n",bytes,l);
        }
    }
    return 0;
}

int closefile_compat() {
    unsigned char channel;
    struct t_buffer *buffer;
    int r;

    if (arguments.mode == M_ETHERNET) {
        channel=ebuf[3] & 0x0f;
        buffer=&buff[channel];
        ebuf[0]=0;ebuf[1]=0;if (sendto(sock,ebuf,2, 0,(struct sockaddr *)&eclient,fromlen)<0) return 1;
    } else {
        r=io_ops->getb();if (r < 0) return 1;
        channel=r & 0x0f;
        buffer=&buff[channel];
        if (io_ops->getb() < 0) return 1;
        if (io_ops->sendb(0)) return 1;
    }
    fprintf(lf,"Close [%d]\n",channel);
    if (buffer->mode==CM_COMPAT) {
        if (buffer->file) fclose(buffer->file);
        buffer->mode=CM_CLOSED;buffer->file=NULL;
        if (buffer->data) {
            free(buffer->data);
            buffer->data=NULL;
        }
    }
    return 0;
}
